# -copyright-
#-# 
#-# xsnow: let it snow on your desktop
#-# Copyright (C) 1984,1988,1990,1993-1995,2000-2001 Rick Jansen
#-# 	      2019,2020,2021,2022 Willem Vermin
#-# 
#-# This program is free software: you can redistribute it and/or modify
#-# it under the terms of the GNU General Public License as published by
#-# the Free Software Foundation, either version 3 of the License, or
#-# (at your option) any later version.
#-# 
#-# This program is distributed in the hope that it will be useful,
#-# but WITHOUT ANY WARRANTY; without even the implied warranty of
#-# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#-# GNU General Public License for more details.
#-# 
#-# You should have received a copy of the GNU General Public License
#-# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#-# 
EXTRA_DIST = \
make.inc \
AltSanta1.xpm \
AltSanta2.xpm \
AltSanta3.xpm \
AltSanta4.xpm \
AltSantaRudolf1.xpm \
AltSantaRudolf2.xpm \
AltSantaRudolf3.xpm \
AltSantaRudolf4.xpm \
BigSanta1.xpm \
BigSanta2.xpm \
BigSanta3.xpm \
BigSanta4.xpm \
BigSanta81.xpm \
BigSanta82.xpm \
BigSanta83.xpm \
BigSanta84.xpm \
BigSantaRudolf1.xpm \
BigSantaRudolf2.xpm \
BigSantaRudolf3.xpm \
BigSantaRudolf4.xpm \
BigSantaRudolf81.xpm \
BigSantaRudolf82.xpm \
BigSantaRudolf83.xpm \
BigSantaRudolf84.xpm \
bird1.xpm \
bird2.xpm \
bird3.xpm \
bird4.xpm \
bird5.xpm \
bird6.xpm \
bird7.xpm \
bird8.xpm \
birdd1.xpm \
birdd2.xpm \
birdd3.xpm \
birdd4.xpm \
birdd5.xpm \
birdd6.xpm \
birdd7.xpm \
birdd8.xpm \
birdl1.xpm \
birdl2.xpm \
birdl3.xpm \
birdl4.xpm \
birdl5.xpm \
birdl6.xpm \
birdl7.xpm \
birdl8.xpm \
eland.xpm \
flake0.xpm \
flake1.xpm \
flake2.xpm \
flake3.xpm \
flake4.xpm \
flake5.xpm \
flake6.xpm \
huis4.xpm \
MediumSanta1.xpm \
MediumSanta2.xpm \
MediumSanta3.xpm \
MediumSanta4.xpm \
MediumSantaRudolf1.xpm \
MediumSantaRudolf2.xpm \
MediumSantaRudolf3.xpm \
MediumSantaRudolf4.xpm \
moon1.xpm \
polarbear.xpm \
RegularSanta1.xpm \
RegularSanta2.xpm \
RegularSanta3.xpm \
RegularSanta4.xpm \
RegularSantaRudolf1.xpm \
RegularSantaRudolf2.xpm \
RegularSantaRudolf3.xpm \
RegularSantaRudolf4.xpm \
rendier.xpm \
snowtree.xpm \
star.xbm \
tannenbaum.xpm \
tree-1_100px.xpm \
tree.xpm \
xsnow41x30.xpm \
xsnow64x64.xpm \
xsnow.svg \
xsnow.xpm 
