#pragma once
/* -copyright-
#-# 
#-# xsnow: let it snow on your desktop
#-# Copyright (C) 1984,1988,1990,1993-1995,2000-2001 Rick Jansen
#-# 	      2019,2020,2021,2022 Willem Vermin
#-# 
#-# This program is free software: you can redistribute it and/or modify
#-# it under the terms of the GNU General Public License as published by
#-# the Free Software Foundation, either version 3 of the License, or
#-# (at your option) any later version.
#-# 
#-# This program is distributed in the hope that it will be useful,
#-# but WITHOUT ANY WARRANTY; without even the implied warranty of
#-# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#-# GNU General Public License for more details.
#-# 
#-# You should have received a copy of the GNU General Public License
#-# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#-# 
*/
#include "../src/Pixmaps/flake0.xpm"
#include "../src/Pixmaps/flake1.xpm"
#include "../src/Pixmaps/flake2.xpm"
#include "../src/Pixmaps/flake3.xpm"
#include "../src/Pixmaps/flake4.xpm"
#include "../src/Pixmaps/flake5.xpm"
#include "../src/Pixmaps/flake6.xpm"
#define SNOW_ALL \
SNOW(0) \
SNOW(1) \
SNOW(2) \
SNOW(3) \
SNOW(4) \
SNOW(5) \
SNOW(6) \

